/***********************************************************************
|
|PROGRAM:   Compton
|FILE:      auxil.c
|AUTHOR:    Marek Gierlinski
|FUNCTIONS: Integral, Idx, Bin, Val, TabulateInverseDistribution, Draw, DrawRnd
|           AddSquares, ComputeErrors, NormalizeSpectra
|GLOBALS:   
|
|
|Version history:
|- started: 8 Feb 1996
|- last update 4 Jul 2008, MG
|
***********************************************************************/
#include "defs.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>



/***************************************************************************
| Integral
|
| Integrates a function using Simpson's rule. From Numerical Recipes.
|
| INPUT:
|   f - a function to be integrated
|   a - lower limit
|   b - upper limit
|   eps - required fractional accuracy
|  
| OUTPUT:
|   Integral - integral of 'f' from 'a' to 'b'
|
| HISTORY:
|   Written: 13 Feb 1996, MG
|   Tested: 14 Feb 1996, MG
***************************************************************************/
#define MAX_ITER 50

double Integral(double (*f)(double), 
		double a, 
		double b,
		double eps)
{
  long int Step, i;
  long int Iter;  /* might be quite big */
  double x, S, Sum, Simp, Delta;
  double OldSum, OldSimp;

  Simp = -1.0e30;
  Sum = 0.5 * (b - a) * ((*f)(a) + (*f)(b));
  Iter = 1;
  Step = 1;

  do
  {
    OldSum = Sum;
    OldSimp = Simp;
    Delta = (b - a) / Iter;
    x = a + 0.5 * Delta;
    S = 0;
    for(i = 1; i <= Iter; i++)
    {
      S += (*f)(x);
      x += Delta;
    }
    Sum = 0.5 * (Sum + (b - a) * S / Iter);
    Simp = (4 * Sum - OldSum) / 3;
    Iter *= 2;
    Step++;
  }
  while(fabs(Simp - OldSimp) > eps * fabs(OldSimp) && Step < MAX_ITER);
  if(Step == MAX_ITER)
    Error("Maximum number of iterations reached in Integral.\n");
  return(Simp);
}

/***************************************************************************
| Idx and Idxm
|
| Index and remainder for a given continuous value and a scale. The returned
| index plus the remainder targets exactly the value of 'Val'. The second
| version (Idxm) does not call error, but returns -1 when out of range.
|
| INPUT:
|   Val - continuous value
|   S - a converting scale
|
| OUTPUT:
|   Idx - discrete index for a value 'Val' and scale 'S'.
|   R - the remainder, Val(Idx) + R * Delta = Val
|
| HISTORY:
|   Written: 12 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
|   Updated: 16 May 1996, MG
***************************************************************************/
int Idxm(double Val,
	 struct Scale S,
	 double *R)
{
  double x;
  int i;
  
  if(S.LogFlag == LOG)
    Val = log10(Val);
  //~ if(Val < S.Min || S.Max < Val)
  //~ {
    //~ if(S.Max < Val && Val < S.Max + 1e-15)       /* case of numerical inacuracy */
      //~ Val = S.Max;
    //~ else
    //~ if(S.Min > Val && Val > S.Min - 1e-15)
      //~ Val = S.Min;
    //~ else
      //~ return(-1);
  //~ }
  if(S.Max < Val)       /* case of numerical inacuracy */
    Val = S.Max;
  else
  if(S.Min > Val)
    Val = S.Min;

  x = (Val - S.Min) / S.Delta;
  i = (int)(x);
  *R = x - i;
  
  if((i < 0) || (i > S.N - 1))
    printf("*** IDX x=%g  i=%d\n", x, i);
  
  return(i);
}

int Idx(double Val,
	struct Scale S,
	double *R)
{
  int i;
    
  i = Idxm(Val, S, R);
  if(i == -1)
  {
    printf("%g %g %g\n", Val, S.Min, S.Max); 
    Error("Value outside range in Idx.\n");
  }
  return(i);
}


/***************************************************************************
| Val
|
| Continuous value for a given index and a scale.
| 
| INPUT:
|   Idx - discrete index
|   S - a converting scale
|
| OUTPUT:
|   Val - countinuous value for an index 'Idx' and scale 'S'
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
***************************************************************************/
double Val(int Idx,
	   struct Scale S)
{
  double v;

  if(Idx < 0 || S.N - 1 < Idx)
    Error("Index outside range in Val.\n");

  if(Idx == S.N - 1)            /* against numerical instability */
    v = S.Max;
  else
    v = S.Min + Idx * S.Delta;

  if(S.LogFlag == LOG)
    return(pow(10.0, v));
  else
    return(v);
}

/***************************************************************************
| Bin
|
| Returns bin (interval) number for a given value
| 
| INPUT:
|   V - given value
|   A - structure with bins
|
| OUTPUT:
|   Bin - bin number
|
| HISTORY:
|   Written: 3 Jul 2008, MG
|   Tested:
***************************************************************************/
int Bin(double V,
	struct Intervals A)
{
  int i;

  if(V < A.X[0] || A.X[A.N] < V)
    Error("Value outside range in Bin.");

  i = 0;
  while(V > A.X[i + 1])
    i++;

  return(i);
}

/***************************************************************************
| DrawRnd
|
| Produces a random value according to a given distribution
|
| INPUT:
|   D - tabulated inverse cumulative distribution
|   Rand - a random number with homogeneous distribution over [0, 1]
|
| OUTPUT:
|   Draw - generated random value
|
| HISTORY:
|   Written: 19 Feb 1996, MG
|   Tested: 19 Apr 1996, MG
***************************************************************************/
double DrawRnd(struct Distribution *D,
	       double Rand)
{
  double x;
  int i;
  double a, b, c;

  if(Rand <= 0.5)
  {
    x = log(Rand * D->x1L0 + 1) / D->logx;
    i = (int)(x);
    if(i == 0)
      i = 1;
  }
  else
  {
    Rand = 1 - Rand;
    x = 2 * D->N - log(Rand * D->x1L0 + 1) / D->logx;
    i = (int)(x);
    if(i == D->N)  /* N - punkt przegiecia, zla aproksymacja */
      i++;
    else if(i == 2 * D->N)
      i--;
  }
  
/* quadratic interpolation y = a x^2 + b x + x */
  
  a = 0.5 * (D->Tab[i - 1] - 2 * D->Tab[i] + D->Tab[i + 1]);
  b = 0.5 * (D->Tab[i + 1] - D->Tab[i - 1]) - 2 * a * i;
  c = D->Tab[i] - a * i * i - b * i;
  return(a * x * x + b * x + c);
}


/***************************************************************************
| Draw
|
| Generates a random value according to a given distribution
|
| INPUT:
|   D - tabulated inverse cumulative distribution
|
| OUTPUT:
|   Draw - generated random value
|
| HISTORY:
|   Written: 19 Feb 1996, MG
|   Tested: 19 Apr 1996, MG
***************************************************************************/
double Draw(struct Distribution *D)
{
  return(DrawRnd(D, Random()));
}


/***************************************************************************
| TabulateInverseDistribution
|
| Tabulates invertion of the cumulative probability distribution.
|
| INPUT:
|   f - probability density distribution function
|   Rng - range over which distribution is calculated
|
| OUTPUT:
|   D - tabulated distribution
|  
| HISTORY:
|   Written: 18 Feb 1996, MG
|   Tested: 19 Apr 1996, MG
***************************************************************************/
#define MAX_DISTR 1000
double FF[MAX_DISTR];   /* Auxiliary array for cumulative distribution */

void TabulateInverseDistribution(double (*f)(double x),
				 struct Range Rng,
				 struct Distribution *D)
{

  int i, L, R;
  double Sum, x, Delta;
  double FL, FR;
  double a, b, c;
  double f0, f1, f2, t0, t1, t2, t;
  
  /* Tabulate cumulative distribution in 'F' */

  FF[0] = 0;
  Sum = 0;
  Delta = (Rng.Max - Rng.Min) / (MAX_DISTR - 1);
  for(i = 1; i <= MAX_DISTR - 1; i++)
  {
    x = Rng.Min + (i - 1) * Delta;
    Sum += Integral(f, x, x + Delta, EPS_DISTRIBUTION);
    FF[i] = Sum;
  }
  
  /* Renormalize distribution to [a, b] interval */
  
  for(i = 1; i <= MAX_DISTR - 1; i++)
    FF[i] /= Sum;
  FF[MAX_DISTR - 1] = 1.0;
  
  
/*  FILE *TT;
  TT = fopen("distr.dat", "wt");
  for(i = 1; i <= MAX_DISTR - 1; i++)
  {
    x = Rng.Min + (i - 1) * Delta;
    fprintf(TT, "%g %g\n", x, FF[i]);
  }
*/   
  
  /* Invert cumulative distribution F into Distribution D */

  D->N = (N_DISTR - 1) / 2;
  D->x = 1.1;
  D->logx = log(D->x);
  D->L0 = 0.5 * (D->x - 1) / (pow(D->x, (float)(D->N)) - 1);
  D->x1L0 = (D->x - 1) / D->L0;

  D->Tab[0] = Rng.Min;
  D->Tab[2 * D->N] = Rng.Max;

  L = 0;
  R = MAX_DISTR - 1;
  for(i = 1; i <= D->N; i++)
  {
    FL = (pow(D->x, (double)i) - 1) / D->x1L0;
    FR = 1 - FL;
    while(FF[L] < FL)
      L++;
    while(FR < FF[R])
      R--;

/* quadratic interpolation t = a f^2 + b f + c, looking for x = t(FL) */    
    
    f0 = FF[L - 1];
    f1 = FF[L];
    f2 = FF[L + 1];
    t0 = Rng.Min + (L - 1) * Delta;
    t1 = t0 + Delta;
    t2 = t1 + Delta;
//    if((f2 - f0) < 0.3)  // arbitrary
//    {
      a = (f0 * t1 + f1 * t2 + f2 * t0 - f0 * t2 - f1 * t0 - f2 * t1)
        / ((f1 - f0) * (f2 - f0) * (f2 - f1));
      b = Delta / (f1 - f0) - a * (f1 + f0);
      c = t0 - a * f0 * f0 - b * f0;
      t = a * FL * FL + b * FL + c;
//    }
//    else 
//    {
//      t = t0 + (FL - f0)/(f1 - f0) * (t1 - t0);
//    }
    D->Tab[i] = t;

    //printf("L=%d i=%d FL=%g f=(%g %g %g) t=(%g %g %g)\n", L, i, FL, f0, f1, f2, t0, t1, t2);    
    //printf("L=%d i=%d FL=%g t=%g\n", L, i, FL, t);    
    
/* quadratic interpolation t = a f^2 + b f + c, looking for x = t(FR) */        

    f0 = FF[R - 1];
    f1 = FF[R];
    f2 = FF[R + 1];
    t0 = Rng.Min + (R - 1) * Delta;
    t1 = t0 + Delta;
    t2 = t1 + Delta;
//    if((f2 - f0) < 0.3)
//    {
      a = (f0 * t1 + f1 * t2 + f2 * t0 - f0 * t2 - f1 * t0 - f2 * t1)
        / ((f1 - f0) * (f2 - f0) * (f2 - f1));
      b = Delta / (f1 - f0) - a * (f1 + f0);
      c = t0 - a * f0 * f0 - b * f0;
      t = a * FR * FR + b * FR + c;
//    }
//    else 
//    {
//      t = t1 + (FR - f1)/(f2 - f1) * (t2 - t1);
//    }
    D->Tab[2 * D->N - i] = t;
    
    //printf("R=%d i=%d FR=%g f=(%g %g %g) t=(%g %g %g) t=%g\n", R, i, FR, f0, f1, f2, t0, t1, t2, t); fflush(stdout);
    
  }
}

/***************************************************************************
| AddSquares
|
| Computes a sum of squares of subspectra. The sum is accumulated in arrays
| TransmittedError and ReflectedError
|
| HISTORY:
|   Written: 8 Jun 1998, MG
|   Tested: 
***************************************************************************/
void AddSquares(int Action)
{
  extern double TransmittedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double TransmittedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern struct Scale SSpectrum;
  extern struct Intervals Angles;

  static double PreviousTrans[MAX_PHOTON_ENERGY][MAX_ANGLE];
  static double PreviousRefl[MAX_PHOTON_ENERGY][MAX_ANGLE];


  int i, j;

  for(j = 0; j <= Angles.N - 1; j++)
    for(i = 0; i <= SSpectrum.N - 1; i++)
      if(Action == INIT)
      {
        PreviousTrans[i][j] = 0.0;
        PreviousRefl[i][j] = 0.0;    
      }
      else
      {
        TransmittedError[i][j] += sqr(TransmittedSpectrum[i][j] - PreviousTrans[i][j]);
        ReflectedError[i][j] += sqr(ReflectedSpectrum[i][j] - PreviousRefl[i][j]);    
        PreviousTrans[i][j] = TransmittedSpectrum[i][j];
        PreviousRefl[i][j] = ReflectedSpectrum[i][j];
      }
}

/***************************************************************************
| ComputeErrors
|
| Computes an estimation of errors. The total spectrum was divided into 
| N_SUBSPECTRA subspectra. Now the standard deviation along these subspectra
| is computed.
|
|
| HISTORY:
|   Written: 8 June 1998, MG
|   Tested: 
***************************************************************************/
void ComputeErrors(void)
{
  extern double TransmittedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double TransmittedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern struct Scale SSpectrum;
  extern struct Intervals Angles;

  int i, j;
   
  for(j = 0; j <= Angles.N - 1; j++)
    for(i = 0; i <= SSpectrum.N - 1; i++)
    {
      TransmittedError[i][j] = sqrt(1.0 / (N_SUBSPECTRA - 1)
              * (TransmittedError[i][j] - sqr(TransmittedSpectrum[i][j]) / N_SUBSPECTRA));
      ReflectedError[i][j] = sqrt(1.0 / (N_SUBSPECTRA - 1)
              * (ReflectedError[i][j] - sqr(ReflectedSpectrum[i][j]) / N_SUBSPECTRA));
    }
}

/***************************************************************************
| NormalizeSpectra
|
| Normalizes the output spectra
|
| INPUT:
|   NumPhoton - number of generated photons
|
| HISTORY:
|   Written: 8 May 1996, MG
|   Tested: 
***************************************************************************/
void NormalizeSpectra(uint64_t NumPhoton)
{
  extern double TransmittedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double TransmittedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double PathDistr[MAX_PATH_DISTR][MAX_ANGLE];
  extern struct Scale SSpectrum, SPathDistr;
  extern struct Intervals Angles;
//  extern int NumPhotons[MAX_ANGLE];

  int i, j;
  double Norm;
  double Val();
   
  for(j = 0; j <= Angles.N - 1; j++)
    for(i = 0; i <= SSpectrum.N - 2; i++)
    {
  
    /* integral norm over the hole spectrum; energy in 'keV' */
  
      Norm = (Val(i + 1, SSpectrum) - Val(i, SSpectrum)) * NumPhoton * MC2;
      Norm *= Angles.X[j + 1] - Angles.X[j];   // normalize to unit area
      TransmittedSpectrum[i][j] /= Norm;
      ReflectedSpectrum[i][j] /= Norm;
      TransmittedError[i][j] /= Norm;
      ReflectedError[i][j] /= Norm;
    }
  
  for(j = 0; j <= Angles.N - 1; j++)
  {
    Norm = 0;
    for(i = 0; i <= SPathDistr.N - 2; i++)
      Norm += (Val(i + 1, SPathDistr) - Val(i, SPathDistr)) * PathDistr[i][j];
    for(i = 0; i <= SPathDistr.N - 2; i++)
      PathDistr[i][j] /= Norm;
  }
}
