/***********************************************************************
|
|PROGRAM:   Compton
|VERSION:   1.27
|FILE:      compton.c
|AUTHOR:    Marek Gierlinski (with minor modifications by Michal Szanecki)
|USES:      defs.h, io.c, init.c, auxil.c, func.c
|FUNCTIONS: main
|GLOBALS:   SeedDistr, SeedFile, ElVelocityDistr, SunyaevTitarchukDistr,
|           CrossSectionTab, TransmittedSpectrum, ReflectedSpectrum,
|           S2Photon, SCross, SSpectrum, RInput, RSpectrum, RElectron,
|           PhotonTemp, ElectronTemp, PhotonEnergyIndex, FoldingEnergy, ElectronEnergyIndex,
|           ElectronSpectrum, OutputUnits
|
|Main file of the program.
|
|The program computes the spectrum of photons resulting from Comptonization 
|of some seed photons off hot electrons. Primary photons are assumed to 
|have a power law or a black body spectrum. They can also have arbitrary
|distrubution created with XSPEC (plot eemo). Electrons have a thermal or
|power-law, homogeneous and isotropic distribution. Electron cloud, of a given
|Thomson depth, might be of plane or spherical geometry.
|
|This is the new, revised and rewritten version of the Gorecki and 
|Wilczewski program, as described in "Acta Astronomica", 1984, 34, 141.
|
|Further description can be found in the enclosed file.
|
|Version history:
|- started: 6 Feb 1996
|- 5 Aug 1996: electron power-law distribution added, MG
|- 17 Oct 1996: minor bug fix, MG
|- 22 Oct 1996: release version 1.0
|- 12 Dec 1996: version 1.01
|               - bug fix: zero spectra for electron kT < 0.6 keV
|-  8 Jun 1998: version 1.02
|               - error estimation of output spectra added
|               - bug fix: improper number of displayed hashes
|- 16 Dec 1998: version 1.03
|               - computing of average no. of scatterings and mean path added
|               - bug fix: generation of electron velocity going out of range
|- 16 Jul 1999: version 1.04
|               - corrected inaccuracies at low energies: increased size 
|                 of auxiliary table 'FF' in 'TabulateInverseDistribution'
|-  7 Oct 1999: version 1.05
|               - seed photons distribution from the central plane in the slab
|                 added
|- 24 Apr 2001: version 1.06
|               - seed, comptonized power and Compton amplification printed
|- 27 Apr 2001: version 1.07
|               - added possibility of excluding unscattered photons
|                 in the spectrum
|-  4 May 2001: version 1.08
|               - fixed bug in writting power law initial spectrum
|- 25 Jun 2008: version 1.1
|               - added arbitraty seed photon distribution from file
|-  4 Jul 2008: version 1.2
|               - photons are now collected in angular bins for slab geometry
|               - display time left, not hashes
|- 31 Jul 2008: version 1.21
|               - minor bug fix
|- 26 Aug 2008: version 1.22
|               - calculating total in-cloud path distribution
|- 18 Sep 2008: version 1.23
|               - NextPosition function rewriten and tidied up
|               - output file formats unified (now energy flux only)
|               - added possibility of neutral absorption in the cloud
|               - number of trajectories is now unsigned long long type
|                 giving max number of 2^64 - 1.
|-  4 Nov 2008: version 1.24 
|               - correction for path outside the sphere in path distribution added
|-  9 Feb 2009: version 1.25
|               - some parameters written in the log file (needed elswhere)
|- 23 Feb 2009: version 1.26
|               - bug in InputData fixed
***********************************************************************/
#include "config.h"
#define DATE            "23 February 2009"

#include "defs.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <errno.h>


/* tabulated probability distributions */                      

struct Distribution BlackBodyDistr;
struct Distribution ElVelocityDistr[MAX_PHOTON_ENERGY];
struct Distribution SunyaevTitarchukDistr;

/* seed photon spectrum from file */

struct PhotonSpectrum SeedFile;

/* tabulated mean cross section */

double CrossSectionTab[MAX_PHOTON_ENERGY];

/* emerging photon counter - arrays for the output spectra */

double TransmittedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
double ReflectedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];

/* arrays for estimating errors -- squares of subspectra */

double TransmittedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
double ReflectedError[MAX_PHOTON_ENERGY][MAX_ANGLE];

/* distribution of total paths */

double PathDistr[MAX_PATH_DISTR][MAX_ANGLE];
double EAvDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza
double E2AvDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza
int NEnDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza

/* scales and ranges of various values */

struct Intervals Angles;
struct Scale S2Photon, SCross, SSpectrum, SPathDistr;
struct Range RInput, RSpectrum, RElectron;

/* main parameters of simulation */

int SeedType;
double PhotonTemp;
double ElectronTemp;
double PhotonEnergyIndex;
double FoldingEnergy;
double ElectronEnergyIndex;
int ElectronSpectrum;
double NonScatteredPhotons;
double AbsorbedPhotons;
double ScatterCounter;
double MeanPath;
double SumWeight;
int NumPhotons[MAX_ANGLE];

/* auxiliary global variables */

int Absorption;
int OutputUnits;
double LogMin, LogMax, PowerLawNorm;

int TEST_FLAG = 0;
uint64_t test_N;

/***************************************************************************
| GenerateInitialPhoton
|
| Generates the initial photon according to appropriate energy and spatial
| distributions.
| 
| INPUT:
|   S - description of the source geometry and initial distribution.
|
| OUTPUT:
|   Ph - generated photon.
|
| HISTORY:
|   Written: 26 Apr 1996, MG
|   Tested: 8 May 1996, MG
|   Updated: 24 May 1996, MG
|   Updated: 16 Dec 1998, MG
|   Updated: 25 Jun 2008, MG
***************************************************************************/
void GenerateInitialPhoton(S, Ph)
struct SourceDescription S;
struct Photon *Ph;
{
  double y;
  double Draw();
  double FileSpectrum();
  void Warning();
  double Integral(), CutoffPowerLaw(), CutoffPowerLawSpectrum();

/* Initial number of scatterings */

  Ph->Scatter = 0;
  Ph->Path = 0.0;
  
/* Generate the energy and statistical weight of the photon */

  switch(S.InitialSpectrum)
    {
    case BLACK_BODY:
      Ph->x = Draw(&BlackBodyDistr);
      Ph->w = 1;
      break;
    case SEED_FILE:
      y = LogMin + Random() * (LogMax - LogMin);
      Ph->x = pow(10.0, y);
      Ph->w = (LogMax - LogMin) * log(10.0) * FileSpectrum(Ph->x) * Ph->x;
      break;
    case POWER_LAW:
      y = LogMin + Random() * (LogMax - LogMin);
      Ph->x = pow(10.0, y);
      Ph->w = pow(Ph->x, -PhotonEnergyIndex) / PowerLawNorm;
      break;
    case E_FOLDED_POWER_LAW:
      y = LogMin + Random() * (LogMax - LogMin);
      Ph->x = pow(10.0, y);
      //      Ph->w = pow(Ph->x, -PhotonEnergyIndex)*exp(-(Ph->x)/FoldingEnergy) / PowerLawNorm;
      Ph->w = CutoffPowerLaw(Ph->x,PhotonEnergyIndex,FoldingEnergy) / PowerLawNorm;
      //      printf("\n wgt= %g %g %g\n", Ph->x,Ph->w,FoldingEnergy);
      break;
    }
  
  Ph->x_init = Ph->x * MC2;
  
/* Generate inclination angle and optical depth of the photon */

  switch(S.Geometry)
  {
    case SLAB:
      switch(S.Distribution)
      {
        case HOMOGENEOUS:
          Ph->tau = Random() * S.tau;
          Ph->eta = 2 * Random() - 1;
          break;
        case CENTRAL:
          Ph->tau = S.tau / 2;
          Ph->eta = 2 * Random() - 1;
          break;
        case CONST_IRRADIATION:
          Ph->tau = S.tau;
          Ph->eta = S.eta0;
          break;
        case THICK_SOURCE:
          Ph->tau = S.tau;
          Ph->eta = sqrt(Random());
          break;
        case THIN_SOURCE:
          Ph->tau = S.tau;
          Ph->eta = Random();
          break;
      }
      break;
    case SPHERE:
      switch(S.Distribution)
      {
        case CENTRAL:
          Ph->tau = 0;
          Ph->eta = 0;
          break;
        case HOMOGENEOUS:
          Ph->tau = S.tau * pow(Random(), 1.0/3);
          Ph->eta = 2 * Random() - 1;
          break;
        case S_T:
          Ph->tau = S.tau * Draw(&SunyaevTitarchukDistr) / PI;
          Ph->eta = 2 * Random() - 1;
          break;
      }
  }
  
  Ph->w_init = Ph->w;
  SumWeight += Ph->w;
}

/***************************************************************************
| NextPosition
|
| Calculates the position of the next scattering and probability of escape.
| 
| INPUT:
|   S - description of the source geometry and initial distribution.
|   Ph - input photon.
|
| OUTPUT:
|   Ph - photon with the new position.
|
| HISTORY:
|   Written: 29 Apr 1996
|   Tested: 10 May 1996, MG
|   Updated: 4 Jul 2008, MG
***************************************************************************/
void NextPosition(
		  struct SourceDescription S,
		  struct Photon *Ph)
{
  double SigmaAbs, SigmaScat;
  double Path, Lambda, NewTau, CosAlpha, PathAngCorrection;
  double ProbabAbs, ProbabEsc, WeightEsc, WeightStay;
  int Angle, Index;
  double Rem;
  double AbsCrossSection(), ScatCrossSection(), PathToEdge(), AbsSigma();
  int Idx(), Idxm(), Bin();
  
  Path = PathToEdge(S, Ph);
  if(Path < 0) return;  /* trapped photon, abort */
  
  
  if(Absorption)
  {
    SigmaAbs = AbsCrossSection(Ph->x);
    ProbabAbs = 1 - exp(-SigmaAbs * Path);
    if(ProbabAbs > Random())
    {
      AbsorbedPhotons += Ph->w;
      Ph->w = 0;                    /* absorbed photon disappears */
      return;
    }
  }
  SigmaScat = ScatCrossSection(Ph->x);
  ProbabEsc = exp(-SigmaScat * Path);
  Lambda = -log(1 - Random() * (1 - ProbabEsc)) / SigmaScat;
  WeightEsc = Ph->w * ProbabEsc;
  WeightStay = Ph->w * (1 - ProbabEsc);

  if(!((S.OutputPhotons == SCATTERED_ONLY) && (Ph->Scatter == 0)))
  {
    Angle = Bin(fabs(Ph->eta), Angles);   /* automatically = 0 for sphere */
    Index = Idx(Ph->x, SSpectrum, &Rem);
    if((S.Geometry == SLAB) && (Ph->eta < 0))       /* going down */
      ReflectedSpectrum[Index][Angle] += WeightEsc;
    else
    {
      TransmittedSpectrum[Index][Angle] += WeightEsc;
      PathAngCorrection = 0;
      if(S.Geometry == SPHERE)
      {
        /* correction for the path outside the sphere */
        CosAlpha = sqrt(1 - Ph->tau*Ph->tau / (S.tau*S.tau) * (1 - Ph->eta*Ph->eta));
        PathAngCorrection = S.tau * (1 - CosAlpha);
      }
      Index = Idxm(Path + Ph->Path + PathAngCorrection, SPathDistr, &Rem);
      if(Index != -1){
	PathDistr[Index][Angle] += WeightEsc;
	EAvDistr[Index][Angle]+=(Ph->x); //misza
	E2AvDistr[Index][Angle]+=((Ph->x)*(Ph->x)); //misza
	NEnDistr[Index][Angle]++; //misza
      //      printf("\n Photon= %i %g\n", Index,Ph->x);
	//	printf("\n Scat= %i %g\n", Ph->Scatter,Ph->x);
      }
    }
  }
  
  ScatterCounter += Ph->Scatter * WeightEsc;
  MeanPath += Path * WeightEsc + Lambda * WeightStay;
  Ph->Path += Lambda;
  if(Ph->Scatter == 0)
    NonScatteredPhotons += WeightEsc;
  Ph->Scatter++;
  Ph->w = WeightStay;
    
  switch(S.Geometry)
  {
    case SPHERE:
      NewTau = sqrt(Ph->tau * Ph->tau + 2 * Ph->tau * Ph->eta * Lambda + Lambda * Lambda);
      Ph->eta = (Lambda + Ph->tau * Ph->eta) / NewTau;
      Ph->tau = NewTau;
      break;
    case SLAB:
      Ph->tau -= Lambda * Ph->eta;
      break;
  }
}

/***************************************************************************
| PathToEdge
|
| Returns distance to the edge of the cloud
| 
| INPUT:
|   S - description of the source geometry and initial distribution.
|   Ph - input photon.
|
| OUTPUT:
|   PathToEdge - distance
|
| HISTORY:
|   Written: 10 Sep 2008, MG
|   Tested:
***************************************************************************/
double PathToEdge(S, Ph)
struct SourceDescription S;
struct Photon *Ph;
{
  double Path;
  double DeltaUp, DeltaDown;
  
  switch(S.Geometry)
  {
    case SPHERE:
      Path = -Ph->tau * Ph->eta 
           + sqrt(S.tau * S.tau - (1 - Ph->eta * Ph->eta) * Ph->tau * Ph->tau);
      break;
    case SLAB:
      DeltaUp = 0.0001 * Ph->tau;    /* +0, -0 to avoid infinite paths */
      DeltaDown = -0.0001 * (S.tau - Ph->tau);
      if(Ph->eta < DeltaDown)
        Path = -(S.tau - Ph->tau) / Ph->eta;  /* down */
      else if(Ph->eta > DeltaUp)
        Path = Ph->tau / Ph->eta;             /* up */
      else
      {
        Ph->w = 0;
        Path = -1;
      }
      break;
  }
  return(Path);
}

/***************************************************************************
| ElectronVelocity
|
| Generates the electron velocity for a given photon energy.
| 
| INPUT:
|   x = h nu / (m c^2) - undimensional photon energy
|
| OUTPUT:
|   El - generated electron velocity
|
| HISTORY:
|   Written: 8 May 1996
|   Tested:
|   Bug fix: 16 Dec 1998
***************************************************************************/
void ElectronVelocity(x, El)
double x;
struct Electron *El;
{
  double gamma, g1, g2;
  double Rand, Rem;
  double TwoXGamma, x1, x2, Fx, Fx1, Fx2, RestX;
  int i;
  void Warning();
  double DrawRnd(), F(), InvF();
  int Idx();
    
  /* interpolation in the electron velocity distributions */
  
  i = Idx(x, S2Photon, &Rem);
  
  /* There is a weird error (probably due to numerical innacuracy at
  the tail of electron velocity distribution), leading to gamma=0,
  so I simply avoid the tail rejecting random numbers very close to 1: */
  
  while((Rand = Random()) > 1 - 1e-10)
    ;
  if(i == S2Photon.N - 1)       /* This means Rem = 0! */
    gamma = DrawRnd(&(ElVelocityDistr[i]), Rand);
  else
  {
    g1 = DrawRnd(&(ElVelocityDistr[i]), Rand);
    g2 = DrawRnd(&(ElVelocityDistr[i + 1]), Rand);
    gamma = g1 + Rem * (g2 - g1);
  }
  if(gamma<1.0) {
      printf("*** Low gamma=%g rounded to 1\n",gamma);
      gamma=1.0;
  }

  El->beta = sqrt(gamma * gamma - 1) / gamma;
  
  /* generate photon energy 'RestX' in the rest frame of the electron */
  
  TwoXGamma = 2 * x * gamma;
  x1 = TwoXGamma * (1 - El->beta);
  x2 = TwoXGamma * (1 + El->beta);
  Fx1 = F(x1);
  Fx2 = F(x2);
  Fx = Fx1 + Random() * (Fx2 - Fx1);
  RestX = InvF(Fx);

  /* correct for potential numerical inacurraces */

  if(RestX > x2) RestX = x2;
  if(RestX < x1) RestX = x1;

  /* now move to the plasma frame and find electron mu = cos(theta) */
  
  El->mu = 1 / El->beta * (1 - RestX / TwoXGamma);

  /* correct for potential numerical inacurraces */

  if(abs(El->mu) >= 1)
  {
/*    Warning("Electron |mu| > 1 in 'ElectronVelocity' function."); */
    El->mu = (El->mu > 1) ? 1 : -1;
  }

  /* electron phi has homogeneous probability distribution */
  
  El->phi = 2 * PI * Random();
}
  
/***************************************************************************
| Scattering
|
| Determine the direction and the energy of the photon after scattering.
| 
| INPUT:
|   El - electron velocity
|   Ph - photon before scattering
|
| OUTPUT:
|   Ph - photon after scattering
|
| HISTORY:
|   Written: 29 Apr 1996
|   Tested:
***************************************************************************/
void Scattering(El, Ph)
struct Electron El;
struct Photon *Ph;
{
  double gamma;
  double PhiP, Mu, MuP, OmegaXP, OmegaZP;
  double CosPhi, CosPhiP, SMu, SMuP;
  double x, xP, xRatio, z, Y;
  double Random2, tmp;
  void Warning();
  double Val();
  
  gamma = 1 / sqrt(1 - El.beta * El.beta);
  Mu = El.mu;
  SMu = sqrt(1 - Mu * Mu);  
  CosPhi = cos(El.phi);
  
  /* photon energy in electron rest frame */
  
  x = 2 * Ph->x * gamma * (1 - El.beta * El.mu);
  
  /* generation of mu' (MuP) and phi' (PhiP) by rejection method */
  
  do
  {
    PhiP = 2 * PI * Random();
    CosPhiP = cos(PhiP);
    
    Random2 = 2 * Random() - 1;
    MuP = (El.beta + Random2) / (1 + El.beta * Random2);
    SMuP = sqrt(1 - MuP * MuP);
    
    OmegaZP = Mu * MuP - SMu * SMuP * CosPhiP;
    
    tmp = 1 - El.beta * MuP;
    xRatio = tmp / (tmp + Ph->x / gamma * (1 - OmegaZP));
    xP = xRatio * x;
    z = 2 / xP - 2 / x;
    Y = xRatio * (1 + xRatio * (xRatio + z * (z - 2)));
  }while(Y < 2 * Random());
    
  /* find the new inclination angle and energy after scattering */
  
  OmegaXP = MuP * CosPhi * SMu 
         + SMuP * (Mu * CosPhi * CosPhiP - sin(El.phi) * sin(PhiP));
  Ph->eta = OmegaZP * Ph->eta - OmegaXP * sqrt(1 - Ph->eta * Ph->eta);
  Ph->x = xP / (2 * gamma * tmp);

  /* correct if photon energy outside range */

  if(Ph->x < RSpectrum.Min)
  {
/*    Warning("Photon energy after scattering < lower spectrum limit."); */
    Ph->x = RSpectrum.Min;
  }
  else if(Ph->x > RSpectrum.Max)
  {
/*    Warning("Photon energy after scattering > upper spectrum limit."); */
    Ph->x = RSpectrum.Max;
  }
}
  




/* Main function */

int *errno_loc;

int main(argc, argv)
int argc;
char *argv[];
{

  void Usage(), ReadData(), CopyData(), InitData(), OpenLogFile();
  void CloseLogFile();
  void TabulateBlackBodyDistr(), TabulateFileDistr(), TabulateElVelocityDistr();
  void TabulateSunyaevTitarchukDistr(), TabulateCrossSection();
  void ComputeErrors(), NormalizeSpectra(), WriteInitialSpectrum(), WriteData();
  void WritePathDistr();
  void Message(), Error(), AddSquares(), ETA(), wr(), wl(), wrlog();
  
  struct InputData Data;
  uint64_t N, DeltaSub;
  struct Photon Ph;
  struct Electron El;
  double pc;
  int NoSquares = 0;
  double TransmittedPower, ReflectedPower, SeedPower;
  time_t start_time, prev_time, now_time;
  
  printf("\n Compton version %s, %s\n\n", VERSION, DATE);
  if(argc != 2)
    Usage();

  errno_loc = __errno_location();
  errno=0;
    
/* Initialization */    
    
  ScatterCounter = 0;
  NonScatteredPhotons = 0;
  AbsorbedPhotons = 0;
  MeanPath = 0;
  SumWeight = 0;

  wr("Initialization...\n");
  wr("  Reading input file");
  ReadData(argv[1], &Data); wl();
  OpenLogFile(Data.LogFile);
  wr("  Initializaing data");
  InitData(Data); wl();
  wr("  Tabulating seed photon distribution");
  if(Data.Source.InitialSpectrum == BLACK_BODY)
    TabulateBlackBodyDistr();
  if(Data.Source.InitialSpectrum == SEED_FILE)
    TabulateFileDistr(Data.SeedFileName);
  wl();
  wr("  Tabulating electron distribution");
  TabulateElVelocityDistr(); wl();
  wr("  Tabulating S-T distribution");
  TabulateSunyaevTitarchukDistr(); wl();
  wr("  Tabulating cross section");
  TabulateCrossSection(); wl();
  AddSquares(INIT);
  wr("  Copying info into log file");
  CopyData(Data); wl();
  wr("Done.\n\n");
  
/* Main program loop */  

#if 0
  /* Not really needed, removed --KL */
  extern unsigned long long int test_N; */
#endif

  Message("Main loop started.");

  start_time = time(NULL);
  prev_time = start_time;
#ifdef DEBUG_MONTE_CARLO
  SRand(1);
#else 
  SRand((long) start_time);  /* initialize random number generator */
#endif
  DeltaSub = Data.NumberTrajectories / N_SUBSPECTRA;
  for(N = 1; N <= Data.NumberTrajectories; N++)
  {
    test_N = N;
    now_time = time(NULL);
    if(now_time > prev_time)
    {
      pc = (double)(N) / Data.NumberTrajectories;
      ETA(start_time, pc);
      prev_time = now_time;
    }
    GenerateInitialPhoton(Data.Source, &Ph);
    while(Ph.w > Data.WeightLimit)
    {
      NextPosition(Data.Source, &Ph);
      if(Ph.w > Data.WeightLimit)
      {
        ElectronVelocity(Ph.x, &El);
        Scattering(El, &Ph);
      }
    }
    if(N - floor(N / DeltaSub) * DeltaSub < 1.0)
    {
      AddSquares(RUN);
      NoSquares++;
    }
  }
  if(NoSquares != N_SUBSPECTRA)
    Error("Internal error SUBNO. Contact the author.");
  printf("\n");
  Message("Main loop finished.");

  ComputeErrors();
  NormalizeSpectra(Data.NumberTrajectories);
  WriteData(Data.TransmittedFile, Data.ReflectedFile, &TransmittedPower, &ReflectedPower);
  WriteInitialSpectrum(Data.Source.InitialSpectrum, Data.InitialFile, &SeedPower);
  WritePathDistr(Data.PathFile);
  wrlog("Average number of scatterings per photon = %g\n", ScatterCounter / Data.NumberTrajectories);
  wrlog("Mean free path of a photon = %g\n", MeanPath / SumWeight);
  wrlog("Fraction of non-scattered escaped photons = %g\n", NonScatteredPhotons / Data.NumberTrajectories);
  wrlog("Fraction of absorbed photons = %g\n", AbsorbedPhotons / Data.NumberTrajectories);
  wrlog("Power in seed photons = %g\n", SeedPower);
  wrlog("Power in transmitted photons = %g\n", TransmittedPower);
  wrlog("Power in reflected photons = %g\n", ReflectedPower);
  Message("\n========Program completed========");
  CloseLogFile();
  return(1);
}
