/***********************************************************************

PROGRAM:   Compton
FILE:      defs.h
AUTHOR:    Marek Gierlinski
FUNCTIONS:
GLOBALS:

File contains global definitions and data structures.

Version history:
- started: 7 Feb 1996
- last update: 30 Mar 2009 -- Krzysio Leszczynski

***********************************************************************/

/* Descriptor identifiers for input data */
#define    _SVID_SOURCE 1


#include "config.h"
#ifdef HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#include <stdlib.h>


#define idNumberTrajectories    1
#define idWeightLimit           2
#define idInputEnergyRange      3
#define idSpectrumEnergyRange   4
#define idElectronEnergyRange   5
#define idPhotonTemperature     6
#define idElectronTemperature   7
#define idPhotonEnergyIndex     8
#define idElectronEnergyIndex   9
#define idThomsonThickness      10
#define idGeometry              11
#define idOutputPhotons         12
#define idDistribution          13
#define idInitialSpectrum       14
#define idElectronSpectrum      15
#define idIrradiationAngle      16
#define idOutputUnits           17
#define idInitialFile           18
#define idTransmittedFile       19
#define idReflectedFile         20
#define idLogFile               21
#define idSeedFile              22
#define idAngularBins           23
#define idPathRange             24
#define idPathFile              25
#define idAbsorption            26
#define idFoldingEnergy         27

#define SLAB                    101
#define SPHERE                  102
#define HOMOGENEOUS             103
#define CENTRAL                 104
#define S_T                     105
#define CONST_IRRADIATION       106
#define THIN_SOURCE             107
#define THICK_SOURCE            108
#define BLACK_BODY              109
#define MAXWELL                 110
#define POWER_LAW               111
#define SEED_FILE               112
#define ALL                     113
#define SCATTERED_ONLY          114
#define mc2                     115
#define keV                     116
#define LOG                     117
#define LIN                     118 
#define YES                     119
#define NO                      120
#define E_FOLDED_POWER_LAW      121

#define INIT                    0
#define RUN                     1

/* Array sizes */

#define MAX_SEED_SPECTRUM       5000
#define MAX_PHOTON_ENERGY       50000
#define MAX_ANGLE               15
#define N_DISTR                 (2 * 500 + 1)
#define MAX_PATH_DISTR          10000

/* Number of subspectra to include in error estimate */

#define N_SUBSPECTRA            5

/* Maximum size of the file name */

#define MAX_FILE_NAME           128

/* Integral accuracies */

#define EPS_DISTRIBUTION        1e-4
#define EPS_ELECTRON            1e-4
#define EPS_CROSS_SECTION       1e-4

/* Geometrical step for arrangement of nodes in the array in Distribution */
/* The field 'x' of the structure 'Distribution' will be set to this value */

#define X_STEP                  1.07

/* Physical and mathematical constants */

#define MC2                     511.0
#ifndef PI
#define PI                      3.14159265358979323846
#endif




#ifdef DOS
#define Random()               ((double)rand()/RAND_MAX)
#define SRand(seed)            (srand((unsigned)seed))
#else
#define Random()               (drand48())
#define SRand(seed)            (srand48((long)seed))
#endif

/* Structure for inverse probability distribution */

struct Distribution
{
  double Tab[N_DISTR];
  int N;
  double x;
  double logx;
  double L0;
  double x1L0;      /* (x - 1) / L0 */
};

/* Description of the cloud and of the source of initial photons */

struct SourceDescription
{
  int Geometry;
  int Distribution;
  int InitialSpectrum;
  int OutputPhotons;
  double tau;
  double eta0;
};

/* Quantity range */

struct Range
{
  double Min, Max;
};

/* Scaling structure for tabulated quantities */

struct Scale
{
  double Min, Max;
  int N;
  double Delta;
  int LogFlag;
};

struct Intervals
{
  double X[MAX_ANGLE];
  int N;
};

/* Photon parameters */

struct Photon
{
  double x;
  double eta;
  double tau;
  double w;
  double x_init;
  double w_init;
  double Path;
  int Scatter;
};

/* Velocity of electron. Angles defined with respect to photon. */

struct Electron
{
  double beta;
  double mu;
  double phi;
};

/* Seed photon spectrum from file */

struct PhotonSpectrum
{
  double Energy[MAX_SEED_SPECTRUM];
  double Flux[MAX_SEED_SPECTRUM];
  int N;
};

/* Structure containg data from the input file */

struct InputData
{
  uint64_t NumberTrajectories;
  double WeightLimit;
  int WritePhotons;
  int Absorption;
  struct Intervals Angles;
  struct Range RInput;
  struct Range RElectron;
  struct Scale SSpectrum;
  struct Scale SPathDistr;
  double PhotonTemp;
  double ElectronTemp;
  double PhotonEnergyIndex;
  double FoldingEnergy;
  double ElectronEnergyIndex;
  int ElectronSpectrum;
  struct SourceDescription Source;
  int OutputUnits;
  char SeedFileName[MAX_FILE_NAME];
  char InitialFile[MAX_FILE_NAME];
  char TransmittedFile[MAX_FILE_NAME];
  char ReflectedFile[MAX_FILE_NAME];
  char PathFile[MAX_FILE_NAME];
  char LogFile[MAX_FILE_NAME];
};



/* Functions, added by KL */
void Error(char *Msg);
void Warning(char *Msg);
void Message(char *Msg);
void TabulateInverseDistribution
     ( double (*f)(double x),
       struct Range Rng,
       struct Distribution *D);


double BlackBodySpectrum(double x);


static inline double sqr(double x)
{
  return(x * x);
}

int Idx(double Val, struct Scale S, double *R);

