/***********************************************************************
|
|PROGRAM:   Compton
|FILE:      func.c
|AUTHOR:    Marek Gierlinski
|FUNCTIONS: sqr, F, InvF, MaxwellDistr, BlackBodySpectrum, ElectronDistr,
|           PhotonDistr, Ro1Prime, SunyaevTitarchuk, ScatCrossSection,
|           AbsCrossSection
|GLOBALS:   PhotonX
|
|
|
|Version history:
|- started: 8 Feb 1996
|- last update 25 Jun 2008 MG
|
***********************************************************************/
#include "defs.h"
#include <math.h>
#include "sigma.h"
#include <stdio.h>
#include <errno.h>

/* Global variables */

double PhotonX;

/***************************************************************************
| sqr
|
| A square
| 
| INPUT:
|   x - a real number
|
| OUTPUT:
|   sqr = x * x
|
| HISTORY:
|   Written: 8 Jun 1998
|   Tested: 
***************************************************************************/
/* Moved to defs.h by KL */

/***************************************************************************
| F
|
| Returns the function F - integral over y * sigma(y), in limits [0, x]
| in fact it returns F0 = F - (17 / 2 - 2 Pi^2 / 3)
| 
| INPUT:
|   x = h nu / (m c^2) - undimensional photon energy
|
| OUTPUT:
|   F - the approximated function
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
|   Rewritten: 31 May 1996, MG
***************************************************************************/
double F(double x) 
{
  double Pos, Rem, V;
  int i;
  
  x = log10(x);
  if(x < F_MIN || x > F_MAX)
    Error("Photon energy outside range in function 'F'.");
  Pos = (x - F_MIN) / F_STEP;
  i = (int)Pos;
  Rem = Pos - i;
  V = FTab[i] + Rem * (FTab[i + 1] - FTab[i]);
  return(pow(10.0, V));
}

/***************************************************************************
| InvF
|
| Approximates the inversion of the function F.
| 
| INPUT:
|   F - value of the function to be inverted.
|
| OUTPUT:
|   InvF - approximated invertion of F.
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
|   Rewritten: 31 May 1996, MG
***************************************************************************/
double InvF(double F)
{
  double Pos, Rem, x;
  int i;
  void Error();
  
  F = log10(F);
  if(F < INVF_MIN || F > INVF_MAX)
    Error("Sigma integral outside range in function 'InvF'.");
  Pos = (F - INVF_MIN) / INVF_STEP;
  i = (int)Pos;
  Rem = Pos - i;
  x = InvFTab[i] + Rem * (InvFTab[i + 1] - InvFTab[i]);
  return(pow(10.0, x));
}  

/***************************************************************************
| MaxwellDistr
|
| Gives the electron Maxwellian distribution (not normalized).
| 
| INPUT:
|   gamma = 1 / Sqrt(1 - (v/c)^2) - electron Lorentz factor
|   Te = (k T) / (m c^2) - undimensional electron temperature.
|
| OUTPUT:
|   MaxwellDistr - Maxwellian distribution for a given 'gamma' and 'Te'.
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested: 22 May 1996 MG
|
| COMMENTS:
|   Factor gamma * sqrt(gamma * gamma - 1) comes from the transformation
|   from the momentum space to the gamma space: 
|     N(gamma) = (mc)^3 gamma Sqrt(gamma - 1) N(p)
|   Additional 1/Te under exp was included to avoid numerical zeros
|   for temperatures Te < 0.001.
***************************************************************************/
double MaxwellDistr(double gamma,
		    double Te) 
{ 
  return(gamma * sqrt(gamma * gamma - 1) * exp(-gamma / Te + 1/Te));
}

/***************************************************************************
| PowerLawDistr
|
| Gives the electron power-law distribution (not normalized)
| N(gamma)=gamma^(-s).
| 
| INPUT:
|   gamma = 1 / Sqrt(1 - (v/c)^2) - electron Lorentz factor
|   s = index of the power-law distribution
|
| OUTPUT:
|   PowerLawDistr - power-law distribution of electrons for the
|   given 'gamma' and 's'
|
| HISTORY:
|   Written: 5 August 1996, MG
|   Tested: 5 August 1996 MG
***************************************************************************/
double PowerLawDistr(double gamma,
		     double s)
{
  return(pow(gamma, -s));
}

/***************************************************************************
| BlackBody
|
| Gives the energetic black body spectrum, normalized to [0, Infinity).
| 
| INPUT:
|   x = (h nu) / (m c^2) - photon energy in m c^2 units,
|   Theta = (k T) / (m c^2) - undimensional photon temperature (T -
|     radiation temperature).
|
| OUTPUT:
|   BlackBody - black body spectrum density for a given energy
|   and temperature.
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
***************************************************************************/
double BlackBody(double x, double Theta)
{
  double xt;
  static double C = 15 / (PI * PI * PI * PI);   /* Normalization */
  
  xt = x / Theta;
  return(C / Theta * xt * xt * xt / (exp(xt) - 1));
}

//misza
double CutoffPowerLaw(double x, double gamma, double ecut)
{
  return(pow(x,-gamma)*exp(-x/ecut));
}

double CutoffPowerLawSpectrum(double x)
{
  extern double PhotonEnergyIndex;
  extern double FoldingEnergy;
  return(CutoffPowerLaw(x,PhotonEnergyIndex,FoldingEnergy) / x);  // photon spectrum
}
//misza

/***************************************************************************
| ElectronDistr
|
| Returns the velocity distribution of electrons. This function is 
| actually used by program.
| 
| INPUT:
|   gamma = 1 / Sqrt(1 - (v/c)^2) - electron Lorentz factor
|
| OUTPUT:
|   ElectronDistr - velocity distribution of electrons.
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Updated: 5 August 1996, MG
|   Tested:
***************************************************************************/
double ElectronDistr(double gamma)
{
  extern double ElectronTemp;
  extern double ElectronEnergyIndex;
  extern int ElectronSpectrum;
  double Distr;
  
  switch(ElectronSpectrum)
  {
    case MAXWELL:
      Distr = MaxwellDistr(gamma, ElectronTemp); 
      break;
    case POWER_LAW:
      Distr = PowerLawDistr(gamma, ElectronEnergyIndex);
      break;
  }  
  return(Distr);
}

/***************************************************************************
| BlackBodyDistr
|
| Returns the spectrum of the initial blackbody photons. The spectrum returned 
| is the photon spectrum, i.e. number of photons of a given energy. This is
| obtained by dividing the oryginal energetic spectrum by energy. This function is 
| actually used by program.
| 
| INPUT:
|   x = (h nu) / (m c^2) - photon energy in m c^2 units,
|
| OUTPUT:
|   BlackBodySpectrum - photon number distribution of initial BB photons.
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
|   Updated: 24 Jun 2008, MG
***************************************************************************/
double BlackBodySpectrum(double x)
{
  extern double PhotonTemp;
 
  return(BlackBody(x, PhotonTemp) / x);  // photon spectrum
}

/***************************************************************************
| FileSpectrum
|
| Uses binary search and interpolation to return the (normalized) seed
| photon distribution from file (SEED). 
| 
| INPUT:
|   x = (h nu) / (m c^2) - photon energy in m c^2 units,
|
| OUTPUT:
|   FileDistr - photon number distribution of initial photons.
|
| HISTORY:
|   Written: 25 Jun 2008, MG
|   Tested:
***************************************************************************/
double FileSpectrum(double x)
{
  extern struct PhotonSpectrum SeedFile;
      
  int left, mid, right;
  double P, frac;
  
  left = 0;
  right = SeedFile.N;
  while(fabs(right - left) > 1)  // binary search
  {
    mid = (int)((left + right) / 2);
    if((SeedFile.Energy[left] - x) * (SeedFile.Energy[mid] - x) > 0)
      left = mid;
    else
      right = mid;
  }
  frac = (x - SeedFile.Energy[left]) 
       / (SeedFile.Energy[right] - SeedFile.Energy[left]);
  P = SeedFile.Flux[left] 
    + (SeedFile.Flux[right] - SeedFile.Flux[left]) * frac;
  return(P);
}
    
/***************************************************************************
| Ro1Prime
|
| Probability distribution for the velocity of electron colliding with the 
| photon. This distribution is NOT normalized.
| 
| INPUT:
|   gamma = 1 / Sqrt(1 - (v/c)^2) - electron Lorentz factor
|   [PhotonX = (h nu) / (m c^2) - photon energy, global variable]
|
| OUTPUT:
|   Ro1Prime - velocity probability distribution
|
| HISTORY:
|   Written: 9 Feb 1996, MG
|   Tested:
***************************************************************************/
double Ro1Prime(double gamma)
{
  double SqGamma;
  double x1, x2;

  SqGamma = sqrt(gamma * gamma - 1);
  x1 = 2 * PhotonX * (gamma - SqGamma);
  x2 = 2 * PhotonX * (gamma + SqGamma);
  return(ElectronDistr(gamma)  * (F(x2) - F(x1)) / (gamma * SqGamma));
}

/***************************************************************************
| SunyaevTitarchuk
|
| Sunyaev-Titarchuk sin-like density probability distribution f(x) = x sin x
| 
| INPUT:
|   x - variable [0..Pi]
|
| OUTPUT:
|   SunyaevTitarchuk - S-T distribution
|
| HISTORY:
|   Written: 26 Apr 1996, MG
|   Tested:
***************************************************************************/
double SunyaevTitarchuk(double x)
{
  return(x * sin(x) / PI);
}

/***************************************************************************
| ScatCrossSection
|
| Returns the approximated mean cross section for scattering, interpolating
| tabulated values in global array 'CrossSectionTab'.
| 
| INPUT:
|   x = h nu / (m c^2) - undimensional photon energy.
|
| OUTPUT:
|   CrossSection - mean cross section.
|
| HISTORY:
|   Written: 12 Feb 1996
|   Tested:
***************************************************************************/
double ScatCrossSection(double x)
{
  extern struct Scale SCross;
  extern double CrossSectionTab[MAX_PHOTON_ENERGY];
  
  double R;
  int Index;
  
  Index = Idx(x, SCross, &R);
  return(CrossSectionTab[Index] +
         R * (CrossSectionTab[Index + 1] - CrossSectionTab[Index]));
}

/***************************************************************************
| AbsCrossSection
|
| Cross section (in sigma_T units) for absorption. 
| Morrison & McCammon 1983, ApJ, 270, 119
| 
| INPUT:
|   x - photon energy in m c^2 units
|
| OUTPUT:
|   AbsCrossSection - sigma(E) / sigma_T
|
| HISTORY:
|   Written: 09/09/2008, MG
|   Tested: 09/09/2008, MG
***************************************************************************/
double AbsCrossSection(double x)
{
  int N = 16;
  double c0[] = {0.336, 0.173, 0.346, 0.781, 0.714, 0.955, 3.089, 1.206, 1.413, 2.027, 3.427, 3.522, 4.339, 6.29, 7.012, 9.532};
  double c1[] = {0.0, 6.081, 2.679, .188, .668, 1.458, -3.806, 1.693, 1.468, 1.047, 0.187, 0.187, -0.024, 0.309, 0.252, 0.0};
  double c2[] = {0.0, -21.5, -4.761, 0.043, -0.514, -0.611, 2.94, -0.477, -0.315, -0.17, 0.0, 0.0, 0.0075, 0.0, 0.0, 0.0};
  double er[] = {0.03, 0.1, 0.284, 0.4, 0.532, 0.707, 0.867, 1.303, 1.840, 2.471, 3.210, 4.038, 7.111, 8.331, 10.0, 1.0e10};
  
  int left, mid, right;
  double sig;
  double sigmaT = 6.6524615e-3;  // sigma_T in units of 10^-22 cm^2
  
  x *= MC2;  // do it in keV
  
  if(x <= 0.03) 
    return(0.336 / (sigmaT * x*x*x));
  if(x >= 10.0)
    return(9.532 / (sigmaT * x*x*x));

  left = 0;
  right = N;
  while(fabs(right - left) > 1)  // binary search
  {
    mid = (int)((left + right) / 2);
    if((er[left] - x) * (er[mid] - x) > 0)
      left = mid;
    else
      right = mid;
  }
  if(er[mid] > x)
    mid--;
  
  mid++;
  sig = (c0[mid] + c1[mid]*x + c2[mid]*x*x) / (x*x*x);
  return(sig / sigmaT);
}

