/***********************************************************************
|
|PROGRAM:   Compton
|FILE:      init.c
|AUTHOR:    Marek Gierlinski
|USES:      defs.h
|USED BY:   compton.c
|FUNCTIONS: Usage, ConvertScale, InitData, TabulateBlackBodyDistr, 
|           TabulateSunyaevTitarchukDistr, TabulateElVelocityDistr,
|           TabulateCrossSection, TabulateFileDistr
|GLOBALS:   
|
|Contains functions initializing global variables: scales, probability 
|distributions, tabulated quantities.
|
|Version history:
|- started: 7 Feb 1996
|- updated 24 May 1996 MG
|- last update: 5 Aug 1996 MG
|- last update: 25 Jun 2008 MG
|- last update: 8 July 2008 MG
|
***********************************************************************/
#include "defs.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>



/***************************************************************************
| Usage
|   Writes short information about how to use program.
| 
| HISTORY:
|   Written: 7 Feb 1996, MG
|   Tested: 7 Feb 1996, MG
***************************************************************************/

void Usage()
{
  printf("\nUsage:\n");
  printf("  compton <input file>\n");
  exit(1);
}

/***************************************************************************
| ConvertScale
|
| Converts a scale to mc^2 units and do logarithms if necessary 
|
| INPUT:
|   S - a scale
|
| OUTPUT:
|   S - converted scale
|
| HISTORY:
|   Written: 8 Feb 1996, MG
|   Tested: 12 Feb 1996, MG
***************************************************************************/
void ConvertScale(struct Scale *S)
{
  S->Min = S->Min / MC2;
  S->Max = S->Max / MC2;
  if(S->LogFlag == LOG)
  {
    S->Min = log10(S->Min);
    S->Max = log10(S->Max);
  }
  S->Delta = (S->Max - S->Min) / (S->N - 1);
}

/***************************************************************************
| InitData
|
| Initialize global scales and other variables
|
| INPUT:
|   Data - structure containing input data
|
| OUTPUT:
|   Global scales and variables
|
| HISTORY:
|   Written: 12 Feb 1996, MG
|   Tested: 14 Feb 1996, MG
|   Updated: 17 May 1996, MG
|   Updated: 24 May 1996, MG
|   Updated: 5 Aug 1996, MG
|   Updated: 17 October 1996, MG
|   Updated: 9 June 1998, MG
|   Updates: 4 July 2008, MG
|   Updates: 23 February 2009, MG
***************************************************************************/
void InitData(struct InputData Data)
{

  /* External variables */
  
  extern double TransmittedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double TransmittedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double PathDistr[MAX_PATH_DISTR][MAX_ANGLE];
  extern double EAvDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza
  extern double E2AvDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza
  extern int NEnDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza 

  extern struct Intervals Angles;
  extern struct Scale S2Photon, SCross, SSpectrum, SPathDistr;
  extern struct Range RInput, RSpectrum, RElectron;
  extern double PhotonEnergyIndex, FoldingEnergy, ElectronEnergyIndex, PhotonTemp, ElectronTemp;
  extern double ScatterCounter, MeanPath;
  extern int ElectronSpectrum;
  extern int OutputUnits;
  extern int Absorption;
  extern double LogMin, LogMax, PowerLawNorm;
  double Integral(), CutoffPowerLaw(), CutoffPowerLawSpectrum();
  
  int i, j;

  /* Check for range errors */
  
  if(Data.SSpectrum.N > MAX_PHOTON_ENERGY)
    Error("Number of spectrum energy points exceeds the maximum.");

  /* Zero the spectra */

  for(j = 0; j <= Data.Angles.N - 1; j++)
  {
    for(i = 0; i < MAX_PHOTON_ENERGY; i++)
    {
      TransmittedSpectrum[i][j] = 0.0;
      ReflectedSpectrum[i][j] = 0.0;
      TransmittedError[i][j] = 0.0;
      ReflectedError[i][j] = 0.0;
    }
    for(i = 0; i < MAX_PATH_DISTR; i++)
      {
      PathDistr[i][j] = 0.0;
      EAvDistr[i][j] = 0.0; //misza
      E2AvDistr[i][j] = 0.0; //misza
      NEnDistr[i][j] = 0; //misza
      }
  }

  /* Photon and electron temperatures, convert from keV to mc^2 units */
  
  PhotonTemp = Data.PhotonTemp / MC2;
  ElectronTemp = Data.ElectronTemp / MC2;
  PhotonEnergyIndex = Data.PhotonEnergyIndex;
  FoldingEnergy = Data.FoldingEnergy / MC2;  
  ElectronEnergyIndex = Data.ElectronEnergyIndex;
  ElectronSpectrum = Data.ElectronSpectrum;
  ScatterCounter = 0;
  MeanPath = 0;

  /* Angular bins. If not slab geometry set one bin only */
  
  if(Data.Source.Geometry == SLAB)
    Angles = Data.Angles;
  else
  {
    Angles.X[0] = 0;
    Angles.X[1] = 1;
    Angles.N = 1;
  }
    
  /* Auxiliary variable, admissible photon energy range during simulation */
  
  RSpectrum.Min = Data.SSpectrum.Min / MC2;
  RSpectrum.Max = Data.SSpectrum.Max / MC2;

  /* Convert photon energies from keV to mc^2 units */

  ConvertScale(&Data.SSpectrum);

  /* Photon energy scale for two-dimensional distribution ro1' */
  
  S2Photon = Data.SSpectrum;

  /* Photon energy scale for tabulated cross section */
  
  SCross = Data.SSpectrum;

  /* Energy scale for emerging photons -- computed spectra */
  
  SSpectrum = Data.SSpectrum;
  
  /* Scale for path distribution */
  
  SPathDistr = Data.SPathDistr;
  if(SPathDistr.LogFlag == LOG)
  {
    SPathDistr.Min = log10(SPathDistr.Min);
    SPathDistr.Max = log10(SPathDistr.Max);
  }
  SPathDistr.Delta = (SPathDistr.Max - SPathDistr.Min) / (SPathDistr.N - 1);
    
  /* Absorption switch */
  
  Absorption = (Data.Absorption == YES);
  
  /* Input photon energy range */
  
  RInput = Data.RInput;
  RInput.Min /= MC2;
  RInput.Max /= MC2;
  if(RInput.Min < RSpectrum.Min || RInput.Max > RSpectrum.Max)
    Error("Input spectrum range exceedes output spectrum energy range.");

  /* Electron gamma scale for two-dimensional distribution ro1' */
  
  RElectron = Data.RElectron;

  /* Convert electron energies from kT units to gamma Lorentz factors */
  /* The formula is gamma = x T + 1 */

  RElectron.Min = RElectron.Min * ElectronTemp + 1;
  RElectron.Max = RElectron.Max * ElectronTemp + 1;
  if(RElectron.Min == 1.0)
    Warning("Electron minimum gamma equals 1.0.");
    
  /* Units of output spectra */
  
  OutputUnits = Data.OutputUnits;
  
  LogMin = log10(RInput.Min);
  LogMax = log10(RInput.Max);
  
  /* Auxiliary variables for power-law generation */
  
  if(Data.Source.InitialSpectrum == POWER_LAW)
  {
    double P = PhotonEnergyIndex;
  
    if(P == 0.0)
      PowerLawNorm = 1;
    else
      //Below was an error? (misza)
      //      PowerLawNorm = pow(RInput.Min, -P) - pow(RInput.Max, -P)
      //	/ (P * (LogMax - LogMin) * log(10.0));
      //Should be?:
      PowerLawNorm = (pow(RInput.Min, -P) - pow(RInput.Max, -P))
      	/ (P * (LogMax - LogMin) * log(10.0));
  }

    if(Data.Source.InitialSpectrum == E_FOLDED_POWER_LAW)
  {
    double P = PhotonEnergyIndex;
    double Efold = FoldingEnergy;

    PowerLawNorm = (pow(RInput.Min, -P) - pow(RInput.Max, -P))
      / (P * (LogMax - LogMin) * log(10.0));
    printf("\n Norm PL= %g \n", PowerLawNorm);
      
    PowerLawNorm = Integral(CutoffPowerLawSpectrum,RInput.Min,RInput.Max,1e-3)
      / ((LogMax - LogMin) * log(10.0));

    printf("\n Norm EPL= %g \n", PowerLawNorm);
    printf("\n %g %g \n", RInput.Min*MC2, RInput.Max*MC2);
      //      pow(1.0/MC2, -P)*exp(-(1.0/MC2)/FoldingEnergy)
      //      / ((LogMax - LogMin) * log(10.0));
  }
}

/***************************************************************************
| TabulateBlackBodyDistr
|
| Tabulates the black body spectrum cumulative distribution.
|
| OUTPUT:
|   BlackBodyDistr - tabulated distribution
|
| HISTORY:
|   Written: 26 Apr 1996, MG
|   Tested:                      
***************************************************************************/
void TabulateBlackBodyDistr(void)
{
  extern struct Range RInput;
  extern struct Distribution BlackBodyDistr;
  extern int SeedType;

  
  SeedType = BLACK_BODY;
  Message("Tabulating black body distribution...");
  TabulateInverseDistribution(BlackBodySpectrum, RInput, &BlackBodyDistr);
#if 0 /* Removed by KL */
  DumpDistribution(SeedDistr);exit(1);
#endif
}

/***************************************************************************
| TabulateFileDistr
|
| Tabulates arbitrary seed photons distribution from file (creates
| cumulative distribution). Uses FileDistr function.
|
| OUTPUT:
|   SeedDistr - tabulated distribution
|
| HISTORY:
|   Written: 23 Jun 2008, MG
|   Updated: 31 Jul 2008, MG
|   Tested:                      
***************************************************************************/
void TabulateFileDistr(char *SeedFileName)
{
  extern struct Range RInput, RSpectrum;
  extern struct PhotonSpectrum SeedFile;
  extern double LogMin, LogMax;

  FILE *S;
  double e, de, f, norm;
  double first_de, last_de;
  int i;
  char dum[1000];

  Message("Tabulating seed photons distribution...");
  S = fopen(SeedFileName, "rt");
  if(S == NULL)
  {
    printf("Cannot open file with seed spectrum %s\n", SeedFileName);
    exit(1);
  }
  
  fgets(dum,1000,S);fgets(dum,1000,S);fgets(dum,1000,S);  /*ignore first 3 qdp lines */
  i = 0;
  while(fscanf(S, "%lf %lf %lf", &e, &de, &f) == 3)
  {
    if(i >= MAX_SEED_SPECTRUM - 1) Error("Too many points in the seed spectrum file.");
    e /= MC2;
    de /= MC2;
    if(i == 0) first_de = de;
    last_de = de;
    SeedFile.Energy[i] = e;
    //    SeedFile.Flux[i] = f;
    SeedFile.Flux[i] = f/e/e; //modified by misza
    norm += f * 2 * de/e/e;
    i++;
    fgets(dum,1000,S); // read the rest of line
  }
  SeedFile.N = i;
  RInput.Min = SeedFile.Energy[0] - first_de;
  RInput.Max = SeedFile.Energy[SeedFile.N - 1] + last_de;
  LogMin = log10(RInput.Min);
  LogMax = log10(RInput.Max);
  if(RInput.Min < RSpectrum.Min || RInput.Max > RSpectrum.Max)
    Error("Seed photon file energy range exceedes output spectrum energy range.");
  
  for(i = 0; i <= SeedFile.N - 1; i++)
    SeedFile.Flux[i] /= norm;             // normalize to integral
}


/***************************************************************************
| TabulateElVelocityDistr
|
| Tabulates a set of electron velocity cumulative distributions for a 
| range of photon energies. 
|
| OUTPUT:
|   ElVelocityDistr - tabulated distribution (global array).
|
| HISTORY:
|   Written: 19 Feb 1996, MG
|   Tested:
***************************************************************************/
void TabulateElVelocityDistr(void)
{
  extern double PhotonX;
  extern struct Range RElectron;
  extern struct Scale S2Photon;
  extern struct Distribution ElVelocityDistr[MAX_PHOTON_ENERGY];

  int i;
    double Ro1Prime(), Val();  
  
  Message("Tabulating electron velocity distribution...");
  if(S2Photon.N > MAX_PHOTON_ENERGY)
    Error("Array size exceeded in function TabulateElVelocityDistr.");
  for(i = 0; i <= S2Photon.N - 1; i++)
  {
    /* PhotonX is a hidden parameter of Ro1Prime function */
    
    PhotonX = Val(i, S2Photon);
    TabulateInverseDistribution(Ro1Prime, RElectron, &(ElVelocityDistr[i]));
  }
}

/***************************************************************************
| TabulateSunyaevTitarchukDistr
|
| Tabulates Sunyaev-Titarchuk cumulative distribution.
|
| OUTPUT:
|   SunyaevTitarchukDistr - tabulated distribution (global structure).
|
| HISTORY:
|   Written: 26 Apr 1996, MG
|   Tested:
***************************************************************************/
void TabulateSunyaevTitarchukDistr()
{
  extern struct Distribution SunyaevTitarchukDistr;
  
  struct Range R = {0.0, PI};
  void Message(), TabulateInverseDistribution();  
  double SunyaevTitarchuk();
  
  Message("Tabulating Sunyaev-Titarchuk distribution...");  
  TabulateInverseDistribution(SunyaevTitarchuk, R, &SunyaevTitarchukDistr);
}

/***************************************************************************
| TabulateCrossSection
|
| Tabulates the mean cross section for Compton scattering in the photon
| energy range [SCross.Min, SCross.Max] (SCross is a global structure).
|
| OUTPUT:
|   CrossSectionTab - tabulated cross section (global array).
|
| HISTORY:
|   Written: 13 Feb 1996, MG
|   Tested: 8 May 1996, MG
***************************************************************************/
void TabulateCrossSection(void)
{
  extern double PhotonX;
  extern struct Range RElectron;
  extern struct Scale SCross;
  extern double CrossSectionTab[MAX_PHOTON_ENERGY]; 

  int i;
  double RoInt, ElInt;
  double gammaMin, gammaMax;
  void Message(), Error();
  double Integral(), ElectronDistr(), Ro1Prime(), Val();  

  Message("Tabulating cross section...");
  if(SCross.N > MAX_PHOTON_ENERGY)
    Error("Array size exceeded in function TabulateCrossSection.");
  gammaMin = RElectron.Min;
  gammaMax = RElectron.Max;

  ElInt =
    (3.0 / 32.0) / Integral(ElectronDistr, gammaMin, gammaMax, EPS_ELECTRON);
    
  for(i = 0; i <= SCross.N - 1; i++)
  {
    /* PhotonX is a hidden parameter of Ro1Prime function */

    PhotonX = Val(i, SCross);
    RoInt = Integral(Ro1Prime, gammaMin, gammaMax, EPS_CROSS_SECTION);
    CrossSectionTab[i] = ElInt * RoInt / (PhotonX * PhotonX);
  }
}


