/***********************************************************************
|
|PROGRAM:   Compton
|FILE:      io.c
|AUTHOR:    Marek Gierlinski
|FUNCTIONS: Error, Warning, Message, Dot, Hash, Percent, ETA, OpenLogFile,
|           CloseLogFile, ReadScale, ReadRange, RecognizeDescriptor, Descriptor,
|           ReadData, CopyData, WriteInitialSpectrum, WriteData
|GLOBALS:   LogFile
|
|Input and output of the program. Functions in this file read the input 
|data from the file and write messages to the log file, and the results to 
|the output file.
|
|Version history:
|- started: 6 Feb 1996
|- last update 4 Jul 2008, MG
|
***********************************************************************/

#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <errno.h>

/* Converting table for descriptors */

struct Conv
{
  char *Descr;
  int ID;
};

static struct Conv ConvTab[] =
{
  {"NumberTrajectories", idNumberTrajectories},
  {"WeightLimit", idWeightLimit},
  {"Absorption", idAbsorption},
  {"AngularBins", idAngularBins},
  {"InputEnergyRange", idInputEnergyRange},
  {"SpectrumEnergyRange", idSpectrumEnergyRange},
  {"ElectronEnergyRange", idElectronEnergyRange},
  {"PathRange", idPathRange},
  {"PhotonTemperature", idPhotonTemperature},
  {"ElectronTemperature", idElectronTemperature},
  {"PhotonEnergyIndex", idPhotonEnergyIndex},
  {"FoldingEnergy", idFoldingEnergy},
  {"ElectronEnergyIndex", idElectronEnergyIndex},
  {"ThomsonThickness", idThomsonThickness},
  {"Geometry", idGeometry},
  {"OutputPhotons", idOutputPhotons},
  {"Distribution", idDistribution},
  {"InitialSpectrum", idInitialSpectrum},
  {"ElectronSpectrum", idElectronSpectrum},
  {"IrradiationAngle", idIrradiationAngle},
  {"OutputUnits", idOutputUnits},
  {"SeedFile", idSeedFile},
  {"InitialFile", idInitialFile},
  {"TransmittedFile", idTransmittedFile},
  {"ReflectedFile", idReflectedFile},
  {"PathFile", idPathFile},
  {"LogFile", idLogFile},
  {"SLAB", SLAB},
  {"SPHERE", SPHERE},
  {"HOMOGENEOUS", HOMOGENEOUS},
  {"CENTRAL", CENTRAL},
  {"S_T", S_T},
  {"CONST_IRRADIATION", CONST_IRRADIATION},
  {"THIN_SOURCE", THIN_SOURCE},
  {"THICK_SOURCE", THICK_SOURCE},
  {"BLACK_BODY", BLACK_BODY},
  {"FILE", SEED_FILE},
  {"MAXWELL", MAXWELL},
  {"POWER_LAW", POWER_LAW},
  {"E_FOLDED_POWER_LAW", E_FOLDED_POWER_LAW},
  {"ALL", ALL},
  {"SCATTERED_ONLY", SCATTERED_ONLY},
  {"mc2", mc2},
  {"keV", keV},
  {"LOG", LOG},
  {"LIN", LIN},
  {"YES", YES},
  {"NO", NO},
  {"***", 0},
};

/* Global variables */

FILE *LogFile;


/***************************************************************************
| Error
|
| The error handling function.
|
| INPUT:
|   Msg - error message to written to the log file
|
| HISTORY:
|   Written: 7 Feb 1996, MG
|   Tested: 7 Feb 1996, MG
|   Updated: 8 May 1996, MG
|   Updated: 22 May 1996, MG
***************************************************************************/
void Error(char *Msg)
{
  printf("\nError: %s\n", Msg);
  if(fprintf(LogFile, "Error: %s\n", Msg) == EOF)
    printf("Cannot write error message to log file: %s\n", Msg);
  else
    fclose(LogFile);
  exit(1);
}

/***************************************************************************
| Warning
|
| Writes the warning to the log file, do not stop the program.
|
| INPUT:
|   Msg - warning message to written to the log file
|
| HISTORY:
|   Written: 18 Feb 1996, MG
|   Tested: 18 Feb 1996, MG
***************************************************************************/
void Warning(char *Msg)
{
  if(fprintf(LogFile, "Warning: %s\n", Msg) == EOF)
    printf("Cannot write error message to log file: %s\n", Msg);
}

/***************************************************************************
| Message
|
| Writes the message to the log file.
|
| INPUT:
|   Msg - message to written to the log file
|
| HISTORY:
|   Written: 18 Feb 1996, MG
|   Tested: 18 Feb 1996, MG
***************************************************************************/
void Message(char *Msg)
{
  if(fprintf(LogFile, "%s\n", Msg) == EOF)
    printf("Cannot write error message to log file: %s\n", Msg);
}

void wr(char *s)
{
  printf("%s", s); fflush(stdout);
}

void wrlog(char *s, double x)
{
  printf(s, x);
  fprintf(LogFile, s, x);
}

void wl()
{
  printf("\n"); fflush(stdout);
}

/***************************************************************************
| Hash
|
| Prints a hash.
|
| HISTORY:
|   Written: 8 Jun 1998, MG
|   Tested: 8 Jun 1998, MG
***************************************************************************/
void Hash(void)
{
  printf("#");
  fflush(stdout);
}

/***************************************************************************
| Percent
|
| Prints percentage
|
| HISTORY:
|   Written: 3 Jul 2008, MG
|   Tested:
***************************************************************************/
void Percent(double p)
{
  p *= 100;
  printf("\b\b\b\b\b%4.1f%%", p);
  fflush(stdout);
}

/***************************************************************************
| ETA
|
| Prints estimated time left
|
| INPUT:
|   t0 - start time
|   frac - fraction of job done
|
| HISTORY:
|   Written: 3 Jul 2008, MG
|   Tested:
***************************************************************************/
void ETA(time_t t0, double frac)
{
  time_t te;
  struct tm *tinfo;
  char buf[80];
  int d;
  
  te = (time_t)((time(NULL) - t0) * (1.0 - frac) / frac);
  tinfo = gmtime(&te);
  strftime(buf, 80, "%H:%M:%S", tinfo);
  if(te < 86400)
    printf(
	"\33[s%4.1f%% done %8s left         \33[u", frac*100, buf);
  else
  {
    d = (int)(te / 86400);
    printf("\33[s%4.1f%% done %3d days %8s left            \33[u", frac*100, d, buf);
  }
  
  fflush(stdout);
}

/***************************************************************************
| OpenLogFile
|
| Opens the log file. Writes error message if error encoutered.
|
| INPUT:
|   FileName - name of the file.
|   
| HISTORY:
|   Written: 6 Feb 1996, MG
|   Tested: 7 Feb 1996, MG
***************************************************************************/
void OpenLogFile(char *FileName)
{
  if((LogFile = fopen(FileName, "wt")) == NULL)
  {
    printf("\nCannot open log file %s.\n", FileName);
    exit(1);
  }
}

/***************************************************************************
| CloseLogFile
|
| Closes the log file. 
|   
| HISTORY:
|   Written: 6 Feb 1996, MG
|   Tested:
***************************************************************************/
void CloseLogFile(void)
{
  fclose(LogFile);
}

/***************************************************************************
| RecognizeDescriptor
|
| Converts string descriptor to a numerical identifier.
|
| INPUT:
|   s - string desriptor
|
| OUTPUT:
|   ConverDescriptor - integer identifier
|
| HISTORY:
|   Written: 6 Feb 1996, MG
|   Tested: 7 Feb 1996, MG
***************************************************************************/
int RecognizeDescriptor(char *s)
{

  int i;

  i = 0;
  while(strcmp(s, ConvTab[i].Descr) != 0 && ConvTab[i].ID != 0)
    i++;
  if(ConvTab[i].ID == 0)
  {
    printf("\nUnknown descriptor in the input file: %s\n\n", s);
    exit(1);
  }                     
  return(ConvTab[i].ID);  
}

/***************************************************************************
| ReadScale
|
| Reads three numbers and a text from the stream F into the structure R.
|
| INPUT:
|   F - input stream, must be open
|
| OUTPUT:
|   S - structure containing read scale
|
| HISTORY:
|   Written: 6 Feb 1996, MG
|   Tested:
***************************************************************************/
int ReadScale(FILE *F, struct Scale *S)
{
  double Min, Max;
  char str[80];
  int N;

  if(fscanf(F, "%lf %lf %d %s", &Min, &Max, &N, str) != 4)
    {
      printf("\nError reading input file.\n");
      exit(1);
    }
  S->Min = Min;
  S->Max = Max;
  S->N = N;
  S->LogFlag = RecognizeDescriptor(str);     /* str = "LIN" or "LOG" */
  return(1);
  
  /* CAUTION: Delta is not set here! */
  
}

/***************************************************************************
| ReadRange
|
| Reads two numbers from the stream F into the structure R.
|
| INPUT:
|   F - input stream, must be open
|
| OUTPUT:
|   R - structure containing read range
|
| HISTORY:
|   Written: 19 Feb 1996, MG
|   Tested:
***************************************************************************/
int ReadRange(FILE *F, struct Range *R)
{
  double Min, Max;

  if(fscanf(F, "%lf %lf", &Min, &Max) != 2)
    {
      printf("\nError reading input file.\n");
      exit(1);
    }
  R->Min = Min;
  R->Max = Max;
  return(1);
}

/***************************************************************************
| ReadSeries
|
| Reads a series of numbers into an array.
|
| INPUT:
|   F - input stream, must be open
|
| OUTPUT:
|   A - struncture with numbers read
|
| HISTORY:
|   Written: 1 Jul 2008, MG
|   Tested:
***************************************************************************/
int ReadSeries(FILE *F, struct Intervals *A) 
{
  char c;
  char num[32];
  double x, lastx;
  int i, n;
  
  n = 0;
  lastx = -1;

  c = '.';
  while(c != '\n')
  {
    while((c = getc(F)) == ' ' || c == '\t')
      ;
    num[0] = c;
    i = 1;
    while((c = getc(F)) != ' ' && c != '\t' &&c != '\n')
      num[i++] = c;
    num[i] = '\0';
    if(sscanf(num, "%lf", &x) != 1)
      Error("Error in angles.");
    if(x < 0 || x > 1)
      Error("cos angle outside [0, 1]");
    if(x <= lastx)
      Error("Angular bins not in order.");
    if(n >= MAX_ANGLE)
      Error("Too many angular bins.");
    A->X[n++] = x;
    lastx = x;
  }
  if(n <= 1) Error("Not enough angles.");
  A->N = n-1;
  return(1);
}
 
/***************************************************************************
| Descriptor
|
| Converts numerical identifier to the string descriptor. Returns "UNKNOWN"
| string if ID not found.
|
| INPUT:
|   ID - descriptor identifier
|
| OUTPUT:
|   Descriptor - text descriptor
|
| HISTORY:
|   Written: 6 Feb 1996, MG
|   Tested: 7 Feb 1996, MG
***************************************************************************/
 char *Descriptor(int ID)
{
  int i;

  i = 0;
  while(ConvTab[i].ID != ID && ConvTab[i].ID != 0)
    i++;
  if(ConvTab[i].ID == 0)
    return("UNKNOWN");
  else
    return(ConvTab[i].Descr);
}
 
/***************************************************************************
| ReadData
|
| Reads the input data.
|
| INPUT:
|   FileName - name of the input file
|
| OUTPUT:
|   Data - structure containing all the input data
|
| HISTORY:
|   Written: 19 Feb 1996, MG
|   Tested: 19 Feb 1996, MG
|   Updated: 8 May 1996, MG
|   Updated: 22 May 1996 MG
|   Updated: 5 Aug 1996 MG
|   Updated: 27 Apr 2001 MG
***************************************************************************/
void ReadData(char *FileName, struct InputData *Data)
{
  FILE *F;
  char c = 1;
  int i;
  char s[80];
  int scan;

  F = fopen(FileName, "rt");
  if(F == NULL)
  {
    printf("\nCannot open input file %s\n", FileName);
    exit(1);
  }

  while(c != EOF)
  {
    /* Omit white characters */
    
    while((c = getc(F)) == ' ' || c == '\t' || c == '\n')
      ;
    if(c == EOF)
      break;
      
    /* Omit comments */
    
    while(c == '#')
    {
      while((c = getc(F)) != '\n' && c != EOF)
        ;
      if(c == '\n')
        c = getc(F);
    }
    
    if(c == '\n')
      while((c = getc(F)) == ' ' || c == '\t' || c == '\n')
        ;
    if(c == EOF)
      break;
      
    /* Read text descriptor into s[] */
    
    s[0] = c;
    i = 1;
    while((c = getc(F)) != '=' && c != EOF)
      if(c != ' ' && c != '\t' && c!='\n')
        s[i++] = c;
    s[i] = '\0';
    if(c == EOF)
    {
      printf("\nError reading input file in: %s", s);
      exit(1);
    }

    /* Recognize descriptor */
    
    switch(RecognizeDescriptor(s))
    {
      case idNumberTrajectories:
        scan = fscanf(F, "%llu", &Data->NumberTrajectories);
        break;
      case idWeightLimit:
        scan = fscanf(F, "%lf", &Data->WeightLimit);
        break;
      case idAbsorption:
        scan = fscanf(F, "%s", s);
        Data->Absorption = RecognizeDescriptor(s);
        break;
      case idAngularBins:
        scan = ReadSeries(F, &Data->Angles);
        break;
      case idInputEnergyRange:
        scan = ReadRange(F, &Data->RInput);
        break;
      case idSpectrumEnergyRange:
        scan = ReadScale(F, &Data->SSpectrum);
        break;
      case idElectronEnergyRange:
        scan = ReadRange(F, &Data->RElectron);
        break;
      case idPathRange:
        scan = ReadScale(F, &Data->SPathDistr);
        break;
      case idPhotonTemperature:
        scan = fscanf(F, "%lf", &Data->PhotonTemp);
        break;
      case idElectronTemperature:
        scan = fscanf(F, "%lf", &Data->ElectronTemp);
        break;
      case idPhotonEnergyIndex:
        scan = fscanf(F, "%lf", &Data->PhotonEnergyIndex);
        break;
      case idFoldingEnergy:
        scan = fscanf(F, "%lf", &Data->FoldingEnergy);
        break;
      case idElectronEnergyIndex:
        scan = fscanf(F, "%lf", &Data->ElectronEnergyIndex);
        break;
      case idThomsonThickness:
        scan = fscanf(F, "%lf", &Data->Source.tau);
        break;
      case idSeedFile:
        scan = fscanf(F, "%s", Data->SeedFileName);
        break;
      case idGeometry:
        scan = fscanf(F, "%s", s);
        Data->Source.Geometry = RecognizeDescriptor(s);
        break;
      case idOutputPhotons:
        scan = fscanf(F, "%s", s);
        Data->Source.OutputPhotons = RecognizeDescriptor(s);
        break;
      case idDistribution:
        scan = fscanf(F, "%s", s);
        Data->Source.Distribution = RecognizeDescriptor(s);
        break;
      case idInitialSpectrum:
        scan = fscanf(F, "%s", s);
        Data->Source.InitialSpectrum = RecognizeDescriptor(s);
        break;
      case idElectronSpectrum:
        scan = fscanf(F, "%s", s);
        Data->ElectronSpectrum = RecognizeDescriptor(s);
        break;
      case idIrradiationAngle:
        scan = fscanf(F, "%lf", &Data->Source.eta0);
        break;
      case idOutputUnits:
        scan = fscanf(F, "%s", s);
        Data->OutputUnits = RecognizeDescriptor(s);
        break;
      case idInitialFile:
        scan = fscanf(F, "%s", Data->InitialFile);
        break;
      case idTransmittedFile:
        scan = fscanf(F, "%s", Data->TransmittedFile);
        break;
      case idReflectedFile:
        scan = fscanf(F, "%s", Data->ReflectedFile);
        break;
      case idPathFile:
        scan = fscanf(F, "%s", Data->PathFile);
        break;
      case idLogFile:
        scan = fscanf(F, "%s", Data->LogFile);
        break;
    }
    if(scan == 0)
    {
      printf("\nError reading input data.\n");
      exit(1);
    }
  }
  fclose(F);
}


/***************************************************************************
| CopyData
|
| Copies the input data to the stream (for verification purposes).
|
| INPUT:
|   Data - structure containing the input data
|   
| OUTPUT:
|   Output is written to the log file.
|
| COMMENTS
|   LogFile must be open, this function does not test it.
|
| HISTORY:
|   Written: 19 Feb 1996, MG
|   Tested: 7 Feb 1996, MG
|   Updated: 8 May 1996, MG
|   Updated: 22 May 1996 MG
|   Updated: 5 August 1996 MG
***************************************************************************/
void CopyData(struct InputData Data) 
{
  extern struct Range RInput;
  
  int i;
  char *a;
  switch(Data.Absorption)
  {
    case YES: a = "yes"; break;
    case NO: a = "no"; break;
    default: a = "unknown";
  }
    
  fprintf(LogFile, "==========Initial parameters===========\n");
  fprintf(LogFile, "Number of trajectories = %llu\n", Data.NumberTrajectories);
  fprintf(LogFile, "Statistical weight limit = %g\n", Data.WeightLimit);
  fprintf(LogFile, "Use absorption: %s\n", a);
  fprintf(LogFile, "%d angular bins:\n", Data.Angles.N);
  for(i = 0; i <= Data.Angles.N; i++)
    fprintf(LogFile, "  %g", Data.Angles.X[i]);
  fprintf(LogFile, "\n");
  fprintf(LogFile, "Input photons energy range:\n");
  fprintf(LogFile, "  Min = %g keV, Max = %g keV\n",
    RInput.Min*MC2, RInput.Max*MC2);
  fprintf(LogFile, "Electron energy range:\n");
  fprintf(LogFile, "  Min = %g kT, Max = %g kT\n",
    Data.RElectron.Min, Data.RElectron.Max);
  fprintf(LogFile, "  Min = %g keV, Max = %g keV\n",
    Data.RElectron.Min * Data.ElectronTemp, Data.RElectron.Max * Data.ElectronTemp);
  fprintf(LogFile, "  gamma_min = %g, gamma_max = %g\n",
    Data.RElectron.Min * Data.ElectronTemp / MC2 + 1, Data.RElectron.Max * Data.ElectronTemp / MC2 + 1);
  fprintf(LogFile, "Output spectrum energy range:\n");
  fprintf(LogFile, "  Min = %g keV, Max = %g keV, N = %d, %s\n",
    Data.SSpectrum.Min, Data.SSpectrum.Max, Data.SSpectrum.N,
    Descriptor(Data.SSpectrum.LogFlag));
  fprintf(LogFile, "Path distribution (optical depth) range:\n");
  fprintf(LogFile, "  Min = %g, Max = %g, N = %d, %s\n",
    Data.SPathDistr.Min, Data.SPathDistr.Max, Data.SPathDistr.N,
    Descriptor(Data.SPathDistr.LogFlag));
  fprintf(LogFile, "Photon temperature = %g keV\n", Data.PhotonTemp);
  fprintf(LogFile, "Electron temperature = %g keV\n", Data.ElectronTemp);
  fprintf(LogFile, "Photon index = %g\n", Data.PhotonEnergyIndex);
  fprintf(LogFile, "Folding energy = %g\n", Data.FoldingEnergy);
  fprintf(LogFile, "Electron index = %g\n", Data.ElectronEnergyIndex);  
  fprintf(LogFile, "Geometry: %s\n", Descriptor(Data.Source.Geometry));
  fprintf(LogFile, "Output photons: %s\n", Descriptor(Data.Source.OutputPhotons));
  fprintf(LogFile, "Distribution: %s\n", Descriptor(Data.Source.Distribution));
  fprintf(LogFile, "Initial spectrum: %s\n", Descriptor(Data.Source.InitialSpectrum));
  fprintf(LogFile, "Electron spectrum: %s\n", Descriptor(Data.ElectronSpectrum));  
  fprintf(LogFile, "Thomson thickness = %g\n", Data.Source.tau);
  fprintf(LogFile, "Irradiation angle = %g\n", Data.Source.eta0);
  fprintf(LogFile, "Output spectrum units: %s\n", Descriptor(Data.OutputUnits));
  fprintf(LogFile, "Seed file: %s\n", Data.SeedFileName);
  fprintf(LogFile, "File with initial spectrum: %s\n", Data.InitialFile);
  fprintf(LogFile, "File with transmitted spectrum: %s\n", Data.TransmittedFile);  
  fprintf(LogFile, "File with reflected spectrum: %s\n", Data.ReflectedFile);
  fprintf(LogFile, "File with path distribution: %s\n", Data.PathFile);
  fprintf(LogFile, "Log file (this file): %s\n", Data.LogFile);
  fprintf(LogFile, "\n");
  fflush(LogFile);
}

/***************************************************************************
| WriteInitialSpectrum
|
| Writes the initial spectrum to the output file
|
| INPUT:
|   Distr - distribution of the spectrum (BLACK_BODY or POWER_LAW)
|   FileName - files name
|
| HISTORY:
|   Written: 22 May 1996, MG
|   Updated: 04 May 2001, MG
|   Tested: 
***************************************************************************/
#define MAX_INIT 100
void WriteInitialSpectrum(int Distr,
			  char *FileName,
			  double *SeedPower)
{
  extern struct Range RInput;
  extern int OutputUnits;
  extern double PowerLawNorm, LogMin, LogMax;
  double Val(), Integral(), BlackBodySpectrum(), FileSpectrum();
  double CutoffPowerLaw();

  FILE *F;
  struct Scale SInput;
  double x, oldx, V, OldV, Norm; 
  extern double PhotonEnergyIndex;
  extern double FoldingEnergy;
  int i; 
  extern int TEST_FLAG;

  F = fopen(FileName, "wt");
  if(F == NULL)
    Error("Cannot open output file with the initial spectrum.");
    
  /* auxiliary logarithmic scale for the input spectrum */

  SInput.Min = log10(RInput.Min);
  SInput.Max = log10(RInput.Max);
  SInput.N = MAX_INIT;
  SInput.LogFlag = LOG;
  SInput.Delta = (SInput.Max - SInput.Min) / (SInput.N - 1);

  *SeedPower = 0;

  switch(Distr)
  {
    case BLACK_BODY:
    case SEED_FILE:
      if(Distr == BLACK_BODY)
        Norm = Integral(BlackBodySpectrum, RInput.Min, RInput.Max, 1e-6) * MC2;
      else
        Norm = 1 * MC2;
      if(Norm == 0)
      {
        Warning("Cannot write initial spectrum due to numerical errors in integration.");
        return;
      }
      TEST_FLAG = 0;
      for(i = 0; i <= SInput.N - 1; i++)
      {
        x = Val(i, SInput);        
        V = (Distr == BLACK_BODY) ? BlackBodySpectrum(x) / Norm : FileSpectrum(x) / Norm;
        if(OutputUnits == keV)
          x *= MC2;
        if(V > 1e-30)
          fprintf(F, "%g  %g\n", x, x * V);
        if(i > 0)
          *SeedPower += 0.5 * (oldx * OldV + x * V) * (x - oldx);
        oldx = x;
        OldV = V;
      }
      break;
    case POWER_LAW:
      Norm = PowerLawNorm * (LogMax - LogMin) * log(10.0) * MC2;
      for(i = 0; i <= SInput.N - 1; i++)
      {
        x = Val(i, SInput);
        V = pow(x, -PhotonEnergyIndex - 1) / Norm;
        if(OutputUnits == keV)
          x *= MC2;
        if(V > 1e-30)
          fprintf(F, "%g  %g  %g  %g\n", x, V, x * V, x * x * V);
        if(i > 0)
          *SeedPower += 0.5 * (oldx * OldV + x * V) * (x - oldx);
        oldx = x;
        OldV = V;
      }
      break;
   case E_FOLDED_POWER_LAW:
      Norm = PowerLawNorm * (LogMax - LogMin) * log(10.0) * MC2;
      for(i = 0; i <= SInput.N - 1; i++)
      {
        x = Val(i, SInput);
	V = CutoffPowerLaw(x,PhotonEnergyIndex+1,FoldingEnergy) / Norm;
	//        V = pow(x, -PhotonEnergyIndex - 1)*exp(-x/FoldingEnergy) / Norm;
        if(OutputUnits == keV)
          x *= MC2;
        if(V > 1e-30)
          fprintf(F, "%g  %g  %g  %g\n", x, V, x * V, x * x * V);
        if(i > 0)
          *SeedPower += 0.5 * (oldx * OldV + x * V) * (x - oldx);
        oldx = x;
        OldV = V;
      }
      break;
  }
  fclose(F);
}
  
/***************************************************************************
| WriteData
|
| Writes the output data.
|
| INPUT:
|   TransFileName, ReflFileName - names of the output files
|
| OUTPUT
|    TransmittedPower, ReflectedPower - integrated energy of the spectra
|
| HISTORY:
|   Written: 22 May 1996, MG
|   Updated: 24 Apr 2001, MG
|   Tested: 
***************************************************************************/
void WriteData(char *TransFileName, char *ReflFileName,
	       double *TransmittedPower, 
	       double *ReflectedPower)
{
  extern struct Scale SSpectrum;
  extern double TransmittedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedSpectrum[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double TransmittedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern double ReflectedError[MAX_PHOTON_ENERGY][MAX_ANGLE];
  extern int OutputUnits;
  extern struct Intervals Angles;

  FILE *FTrans, *FRefl;
  int i, j;
  double x, dx;
  double ST, SR, ET, ER;
  double Val();

  FTrans = fopen(TransFileName, "wt");
  if(FTrans == NULL)
    Error("Cannot open output file with transmitted spectrum.");
  FRefl = fopen(ReflFileName, "wt");
  if(FRefl == NULL)
    Error("Cannot open output file with reflected spectrum.");

  *TransmittedPower = 0;
  *ReflectedPower = 0;
  for(i = 0; i <= SSpectrum.N - 2; i++)
  {
    /* x - middle energy of the bin */

    x = (Val(i, SSpectrum) + Val(i + 1, SSpectrum)) / 2;
    dx =  (Val(i+1, SSpectrum) - Val(i, SSpectrum));
    if(OutputUnits == keV)
    {
      x *= MC2;
      dx *= MC2;
    }
    fprintf(FTrans, "%g ", x);
    fprintf(FRefl, "%g ", x);
    for(j = 0; j <= Angles.N - 1; j++)
    {
      ST = TransmittedSpectrum[i][j];
      SR = ReflectedSpectrum[i][j];
      ET = TransmittedError[i][j];
      ER = ReflectedError[i][j];
      *TransmittedPower += x * ST * dx;
      *ReflectedPower += x * SR * dx;
//      if(ST <= 0)
//        ST = 1e-32;
//      fprintf(FTrans, "  %g  %g", x * ST, x * ET);
      fprintf(FTrans, "  %g", x * x * ST); //modified by misza
//      if(SR <= 0)
//        SR = 1e-32;
//      fprintf(FRefl, "  %g  %g", x * SR, x * ER);
      fprintf(FRefl, "  %g", x * x* SR); //modified by misza
    }
    fprintf(FTrans, "\n");
    fprintf(FRefl, "\n");
  }
  fclose(FTrans);
  fclose(FRefl);
}


/***************************************************************************
| WritePathDistr
|
| Writes path distribution
|
| INPUT:
|   
|
| OUTPUT
|    
|
| HISTORY:
|   Written:
|   Updated:
|   Tested: 
***************************************************************************/
void WritePathDistr(char *PathFileName)
{
  extern struct Scale SPathDistr;
  extern double PathDistr[MAX_PATH_DISTR][MAX_ANGLE];
  extern double EAvDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza
  extern double E2AvDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza
  extern int NEnDistr[MAX_PATH_DISTR][MAX_ANGLE]; //misza 
  extern struct Intervals Angles;

  FILE *F;
  int i, j;
  double t, dt;
  double Val();

  F = fopen(PathFileName, "wt");
  if(F == NULL)
    Error("Cannot open output file with path distribution.");

  for(i = 0; i <= SPathDistr.N - 2; i++)
  {
    t = (Val(i, SPathDistr) + Val(i + 1, SPathDistr)) / 2;
    dt =  (Val(i+1, SPathDistr) - Val(i, SPathDistr));
    fprintf(F, "%g ", t);
    for(j = 0; j <= Angles.N - 1; j++)
      {
	if(NEnDistr[i][j]>0) //misza
	  {
	    EAvDistr[i][j]/=NEnDistr[i][j]; //misza
	    E2AvDistr[i][j]/=NEnDistr[i][j]; //misza
	  }
	fprintf(F, "  %g %g %g", PathDistr[i][j],EAvDistr[i][j],E2AvDistr[i][j]);

	//	printf("\n %i %i %i \n", i , j, NEnDistr[i][j]); //misza
      }
    fprintf(F, "\n");
  }
  fclose(F);
}





// For dubugging purposes

void DumpDistribution(char *File, struct Distribution D)
{
  int i;
  FILE *T;
  
  T = fopen(File, "wt");
  

  for(i = 0; i <= 2 * D.N; i++)
    fprintf(T, "%d %g\n", i, D.Tab[i]);
}
